# This file is part of the "FDN tools" project.
# You can use and redistribute it under the terms of the
# GNU General Public License.
# (c) French Data Network 1997-2005
# Authors and contributors:
# - Benjamin Bayart
# - Stephane Jourdois
package FDN::Common::Remember;
#
# Remember
# - Implmentation de smaphores multi-hosts
# - Consigne de data pour processus
# - Fait le caf aussi (voir avec benj pour le matriel)
# $Id: Remember.pm,v 1.3 2010-08-18 23:15:53 benjamin Exp $

use strict;
use Exporter ;
use IO::Handle;
use Net::Domain qw( hostname );
use Storable;
use FDN::Common::Logger;
use FDN::Common::FileLock;
use FDN::Common::Constantes;
use Fcntl;

our @ISA = ('Exporter');
our @EXPORT = qw(
	&process_start
	&process_stop
	&process_abort

	&process_get_latest
	&process_set_latest
	&process_get_deposit
	&process_set_deposit
	&process_get_current
);


# Ce rpertoire _est_ sur le filer, sinon tout a n'a aucun intret.
our $REP = "/var/lib/sitalibs/remember";

# Les suffixes  ajouter  l'id (le prfixe tant $REP) pour obtenir
# un nom de fichier unique sur le filer
my $LOCKSUFFIX = ".lock";
my $DATASUFFIX = ".data";
my $LASTSUFFIX = ".latest";

# Cette structure contiendra tous les filehandles dont cette lib peut
# avoir besoin, pour ne pas polluer
my %fhs;

# Ce hash contient toutes les infos disponibles sur le dernier process
# qui s'est termin avec succs (la cl est l'id, toujours).
my %latest;

# Ce hash contient toutes les infos disponibles sur le process courant
# (rempli par process_start)
my %current;

# Cette fonction essaye de poser un lock pour l'id donn.
# Elle renvoie 1 ou 0 selon qu'un lock a t pos avec succs.
# Si le second paramtre est vrai, cette fonction bloque en attendant
#    le lock au lieu de dire non.
# Si des infos sont disponibles sur le dernier lancement russi,
#    elle seront retournes par process_get_latest()).
sub process_start ($;$) {
	my ($id, $wait) = @_;
	start \@_;

	$fhs{$id}{'lock'} = new IO::Handle;

	# Le fh doit tre ouvert en criture pour pouvoir poser le lock
	unless (
		sysopen $fhs{$id}{'lock'},
			"$REP/$id".$LOCKSUFFIX,
			O_RDWR | O_CREAT
	) {
		alert "Ne peut ouvrir $REP/$id$LOCKSUFFIX : $!";
		print STDERR "Cannot open $REP/$id$LOCKSUFFIX : $!\n";
		exit 1;
	}
	# WARNING :
	# le O_RDWR est important (pas O_WRONLY) : il ne faut pas toucher
	#  ce fichier tant qu'on n'a pas de lock dessus !
	# P.S. :
	# on pourrait ici lire ce que contenait ce fichier, mais a
	# n'a pas d'importance. Si ce fichier contient une info sans
	# qu'un lock ne soit pos, c'est qu'il y a eu un exit sauvage
	# rcemment (donc c'est pas tellement le pb de la lib, sauf
	# ventuellement pour dbugguer).

	if (setlock($fhs{$id}{'lock'}, $wait)) {
		# Le processus courant est donc l'lu, que cela soit dit !
		notice "Obtenu le lock pour $id";

		$current{$id}{'hostname'} = hostname();
		$current{$id}{'pid'} = $$;
		$current{$id}{'start'} = time();
		$fhs{$id}{'lock'}->truncate(0);
		$fhs{$id}{'lock'}->print(
			$current{$id}{'hostname'}, "\n",
			$current{$id}{'pid'}, "\n",
			$current{$id}{'start'}, "\n"
		);

		# go, Go, GO !
		return 1;
	} else {
		notice "Le lock pour $id est dj pris (" . (-(-$!)) . ":$!) !";
		$fhs{$id}{'lock'}->close;
		undef $fhs{$id};
		return 0;
	}
}

# Cette fonction stoque les infos de session et de consigne pour l'id
#   donn, puis retire le lock.
# Elle renvoie 1 ou 0 en fonction du succs
#    (ou meurt si maurice pousse le bouchon un peu trop loin).
sub process_stop ($) {
	my ($id) = @_;
	start \@_;

	unless ($fhs{$id} and $fhs{$id}{'lock'} and $fhs{$id}{'lock'}->opened) {
		error "Stop sans start ?";
		die "Stop sans start ?";
	}
	
	# On se souviendra de toi, cher process...
	# Attention, on se sert de $current{$id} parce qu'on est surs de ce
	#    qu'on fait (readonly), ce n'est pas recommand.
	#    voir process_get_current() pour une version sure.
	process_set_latest($id, $current{$id});

	# Merci d'avoir choisi Remember pour effectuer votre tache,
	#   nous esprons vous revoir bientt sur nos lignes (de code).
	if (unsetlock($fhs{$id}{'lock'})) {
		notice "Lock releas pour $id";
		$fhs{$id}{'lock'}->close;
		undef $fhs{$id};
		undef $latest{$id};
		return 1;
	} else {
		error "Lock _NON_ releas pour $id ($!)";
		# FIXME : grer a un peu mieux...
		return 0;
	}
}

# Cette fonction
# - enlve le lock s'il y en a un,
# - abandonne toutes les infos stoques.
# Elle ne peut pas chouer, donc elle renvoie toujours 1.
sub process_abort ($) {
	my ($id) = @_;
	start \@_;
	
	unless ($fhs{$id} and $fhs{$id}{'lock'} and $fhs{$id}{'lock'}->opened) {
		error "Stop sans start ?";
		die "Stop sans start ?";
	}

	if (unsetlock($fhs{$id}{'lock'})) {
		notice "Lock releas pour $id";
		$fhs{$id}{'lock'}->close;
		undef $fhs{$id};
		undef $latest{$id};
		return 1;
	} else {
		error "Lock _NON_ releas pour $id ($!)";
		# FIXME : grer a un peu mieux...
		return 0;
	}
}

# Cette fonction renvoie un rhash avec les infos dispo pour l'id donn.
#    - hostname : le hostname (_pas_ le fqdn),
#    - pid      : le pid... mme si a n'a aucun intret
#    - start    : secondes depuis epoch au moment du start
#    - end      : secondes depuis epoch au moment du end
# Si aucune info n'est dispo, renvoie undef.
# note : cette fonction stoque la valeur dans un cache qui est vid
#        par process_stop et process_abort
sub process_get_latest ($) {
	my ($id) = @_;
	start \@_;
	my ($hostname, $pid, $start, $end);

	unless (defined $latest{$id}) {
		$fhs{$id}{'last'} = new IO::Handle;
		if (open $fhs{$id}{'last'}, "<", "$REP/$id".$LASTSUFFIX) {
			chop($hostname = $fhs{$id}{'last'}->getline);
			chop($pid      = $fhs{$id}{'last'}->getline);
			chop($start    = $fhs{$id}{'last'}->getline);
			chop($end      = $fhs{$id}{'last'}->getline);
			# (bouh que c'est laid !)
			$fhs{$id}{'last'}->close;
			# On stoque et on renvoie
			return $latest{$id} = {
				hostname => $hostname,
				pid	 => $pid,
				start    => $start,
				end      => $end,
			};
		} else {
			return undef;
		}
	}

	# Si on arrive l c'est qu'il faut renvoyer quelque-chose
	# On copie pour ne pas renvoyer un pointeur (readwrite)
	my %tmp;
	while (my ($k, $v) = each %{$latest{$id}}) {
		$tmp{$k} = $v;
	}
	return \%tmp;
}

# L'inverse de get.
# Renvoie 1 ou 0 en fonction du succs.
# rinfos est une rfrence sur un hash avec comme cls :
#     - hostname
#     - pid
#     - start
#     - end (si end est non dfini, il sera dfini  time() avant
# 	d'tre enregistr.
sub process_set_latest ($$) {
	my ($id, $rinfos) = @_;
	start \@_;

	# Si l'heure de fin n'a pas encore t fixe, il est temps
	# de le faire, sinon on garde les infos qu'on nous a donn
	$rinfos->{'end'} = time() unless defined $rinfos->{'end'};

	# On met  jour le cache
	$latest{$id} = $rinfos;

	$fhs{$id}{'last'} = new IO::Handle;
	if (open $fhs{$id}{'last'}, ">", "$REP/$id".$LASTSUFFIX) {
		$fhs{$id}{'last'}->print(
			$rinfos->{'hostname'}, "\n",
			$rinfos->{'pid'}, "\n",
			$rinfos->{'start'}, "\n",
			$rinfos->{'end'}, "\n"
		);
		$fhs{$id}{'last'}->close;
		return 1;
	} else { return 0 }
}

# Cette fonction renvoie les data stoques pour cet id s'il y en a,
# undef sinon
sub process_get_deposit ($) {
	my ($id) = @_;
	start \@_;

	my $filename = "$REP/$id".$DATASUFFIX;
	return retrieve $filename if -s $filename;
	return undef;
}

# Cette fonction stoque les data donnes pour cet id, en crasant
# les infos prcdentes, et sans rien toucher d'autre.
# Elle renvoie 1 en cas de succs, 0 sinon
sub process_set_deposit ($$) {
	my ($id, $rdata) = @_;
	start \@_;

	return store $rdata, "$REP/$id".$DATASUFFIX if defined $rdata;
}

# Cette fonction renvoie un rhash avec les mmes cls que
# process_get_latest, mais pour le process en cours (moi-mme).
sub process_get_current ($) {
	my ($id) = @_;
	start \@_;

	# Si les infos ne sont pas dispo, on renvoie undef
	return undef unless $current{$id};

	# on copie pour ne pas renvoyer une pointeur (readwrite)
	my %tmp;
	while (my ($k, $v) = each %{$current{$id}}) {
		$tmp{$k} = $v;
	}
	return \%tmp;
}

# Il n'y a pas de process_set_current parce que seul process_start est
# certain d'avoir ouvert le fichier avec le bon lock.

1;
